// DCC Headlight Decoder with sound for DS-DCC decode
// By yaasan
// Based on Nicolas's sketch http://blog.nicolas.cx
// Inspired by Geoff Bunza and his 17 Function DCC Decoder & updated library
//
//
// Debug serial output available on the serial port at baud 115200, aka Tools -> Serial Monitor
//

//#define __DEBUG   // シリアルポートデバックフラグ

#include <NmraDcc.h>
#include <stdint.h>
#include <avr/interrupt.h>
#include <avr/io.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>	 //required by notifyCVRead() function if enabled below

#include "imaadpcm.h"
#include "E233_Alert1_imaADPCM.h"
#include "E233_Alert3_imaADPCM.h"
#include "E233_break1_imaADPCM.h"

#define SAMPLE_RATE 8000

//各種設定、宣言

#define DECODER_ADDRESS 3
#define DCC_ACK_PIN 9	//if defined enables the ACK pin functionality. Comment out to disable.

#define MOTOR_PWM_A 9
#define MOTOR_PWM_B 10


#define MAX_PWMDUTY 200
#define MID_PWMDUTY 100

//使用クラスの宣言
NmraDcc	 Dcc;
DCC_MSG	 Packet;

//Task Schedule
unsigned long gPreviousL5 = 0;

//進行方向
uint8_t gDirection = 128;

//Function State
uint8_t gState_F0 = 0;
uint8_t gState_F1 = 0;
uint8_t gState_F2 = 0;
uint8_t gState_F3 = 0;
uint8_t gState_F4 = 0;

//モータ制御関連の変数
uint32_t gSpeedRef = 1;

//CV related
uint8_t gCV1_SAddr = 3;
uint8_t gCVx_LAddr = 3;


//Sound played flag
uint8_t gSound_played = 0;
uint8_t gSound_playnext = 0;
int gSound_Length = 0;
uint8_t SoundState = 0;			// 0: idle  1:Sound再生待ち

int ledPin = 13;
int speakerPin = 3; // Can be either 3 or 11, two PWM outputs connected to Timer 2  ## デコーダー開発ボードは 3

/* ADPCM related */
volatile uint16_t sample;
uint16_t gCntBlockIn;       // 11/9 block数をカウント
byte lastSample;
IMADEC gBufIMAdata;

//IMA-ADPCM  add aya
int gSound_offset = 0;	// aya add  RIFF フォーマットをそのまま使っているのでdata chunkまでoffsetを履かせる

uint8_t pv04_FuncState;              // 2015/9/12 aya add

//Internal variables and other.
#if defined(DCC_ACK_PIN)
const int DccAckPin = DCC_ACK_PIN ;
#endif

struct CVPair{
  uint16_t	CV;
  uint8_t	Value;
};
CVPair FactoryDefaultCVs [] = {
  {CV_MULTIFUNCTION_PRIMARY_ADDRESS, DECODER_ADDRESS},
  {CV_ACCESSORY_DECODER_ADDRESS_MSB, 0},		//The LSB is set CV 1 in the libraries .h file, which is the regular address location, so by setting the MSB to 0 we tell the library to use the same address as the primary address. 0 DECODER_ADDRESS
  {CV_MULTIFUNCTION_EXTENDED_ADDRESS_MSB, 0},	 //XX in the XXYY address
  {CV_MULTIFUNCTION_EXTENDED_ADDRESS_LSB, 0},	 //YY in the XXYY address
  {CV_29_CONFIG, 128 },	 //Make sure this is 0 or else it will be random based on what is in the eeprom which could caue headaches
};

void(* resetFunc) (void) = 0;  //declare reset function at address 0


uint8_t FactoryDefaultCVIndex = sizeof(FactoryDefaultCVs) / sizeof(CVPair);
byte readSoundRom(uint8_t inNo, uint16_t inOffset);


void notifyCVResetFactoryDefault()
{
	//When anything is writen to CV8 reset to defaults. 

	resetCVToDefault();	 
	Serial.println("Resetting...");
	delay(1000);  //typical CV programming sends the same command multiple times - specially since we dont ACK. so ignore them by delaying

	resetFunc();
};


void resetCVToDefault()
{
	//CVをデフォルトにリセット
	Serial.println("CVs being reset to factory defaults");
	
	for (int j=0; j < FactoryDefaultCVIndex; j++ ){
		Dcc.setCV( FactoryDefaultCVs[j].CV, FactoryDefaultCVs[j].Value);
	}
};

void notifyCVAck(void)
{
  Serial.println("notifyCVAck");
  
  analogWrite(MOTOR_PWM_B, 0);
  analogWrite(MOTOR_PWM_A, 64);
  
  delay( 6 );  
  
  analogWrite(MOTOR_PWM_A, 0);
}

extern void	   notifyCVChange( uint16_t CV, uint8_t Value){
   //CVが変更されたときのメッセージ
   Serial.print("CV "); 
   Serial.print(CV); 
   Serial.print(" Changed to "); 
   Serial.println(Value, DEC);
};


void setup()
{
	uint8_t cv_value;
	
	//シリアル通信開始
	Serial.begin(115200);

	//D9,D10 PWM キャリア周期:31kHz
	TCCR1B &= B11111000;
	TCCR1B |= B00000001;
	
	//PWMを10bit化
	//TCCR1A |= B00000011;
	
	pinMode(ledPin, OUTPUT);
	digitalWrite(ledPin, HIGH);

	
	//PWM出力ピン D9,D10を出力にセット
	pinMode(9, OUTPUT);
	pinMode(10, OUTPUT);

	
	//DCCの応答用負荷ピン
	
	#if defined(DCCACKPIN)
	//Setup ACK Pin
	pinMode(DccAckPin,OUTPUT);
	digitalWrite(DccAckPin, 0);
	#endif
  
 

  #if !defined(DECODER_DONT_DEFAULT_CV_ON_POWERUP)
	if ( Dcc.getCV(CV_MULTIFUNCTION_PRIMARY_ADDRESS) == 0xFF ){	 //if eeprom has 0xFF then assume it needs to be programmed
	  Serial.println("CV Defaulting due to blank eeprom");
	  notifyCVResetFactoryDefault();
	  
   } else{
	 Serial.println("CV Not Defaulting");
   }
  #else
	 Serial.println("CV Defaulting Always On Powerup");
	 notifyCVResetFactoryDefault();
  #endif 
  
   
   
  // Setup which External Interrupt, the Pin it's associated with that we're using, disable pullup.
  Dcc.pin(0, 2, 0);
  
  // Call the main DCC Init function to enable the DCC Receiver
  Dcc.init( MAN_ID_DIY, 100,   FLAGS_MY_ADDRESS_ONLY , 0 ); 

  //Reset task
  gPreviousL5 = millis();
  
  //Init CVs
  /*
  gCV1_SAddr = Dcc.getCV( CV_MULTIFUNCTION_PRIMARY_ADDRESS ) ;
  gCVx_LAddr = (Dcc.getCV( CV_MULTIFUNCTION_EXTENDED_ADDRESS_MSB ) << 8) + Dcc.getCV( CV_MULTIFUNCTION_EXTENDED_ADDRESS_LSB );
  
  Serial.print("CV1(ShortAddr): ");
  Serial.println(gCV1_SAddr);
  Serial.print("CV17/18(LongAddr): ");
  Serial.println(gCVx_LAddr);
 
*/  Serial.println("Ready");
 
	 
}

void loop(){
	
	// You MUST call the NmraDcc.process() method frequently from the Arduino loop() function for correct library operation
	Dcc.process();
	

	if( (millis() - gPreviousL5) >= 100)
	{
		//Headlight control
		HeadLight_Control();
		
		//Reset task
		gPreviousL5 = millis();
		
	}
}
//------------------------------------------------------------------------------
//	Motor control Task (10Hz) */
//------------------------------------------------------------------------------
void HeadLight_Control()
{
/*
Serial.print("HeadLight_Control():");
Serial.print("gState_F0:");Serial.print(gState_F0);
Serial.print(",gState_F1:");Serial.print(gState_F1);
Serial.print(",gState_F2:");Serial.print(gState_F2);
Serial.print(",gState_F3:");Serial.print(gState_F3);
Serial.print(",gState_F4:");Serial.println(gState_F4);
*/

	uint16_t aPwmRef = 0;
	
	if( gState_F0 > 0)
	{
		aPwmRef = HIGH;
	}
	else
	{
		aPwmRef = LOW;
	}
	
	
	//PWM出力
	//進行方向でPWMのABを切り替える
	if( gDirection > 0){
		digitalWrite(MOTOR_PWM_B, LOW);
		digitalWrite(MOTOR_PWM_A, aPwmRef);
	}
	else
	{
		digitalWrite(MOTOR_PWM_A, LOW);
		digitalWrite(MOTOR_PWM_B, aPwmRef);
	}
	
	//再生が終わっていたらここで関数を完了する
	if( gSound_played != 0)
	{
		return;
	}
	
	if((SoundState == 1) && (gSound_playnext == 1))	// DCS50Kは常にファンクッションコマンド送り続けるため、そのガード。
//	if( (gState_F2 > 0) && (gSound_playnext == 1))
	{
		//Horn (Long)
		gSound_played = 1;
		
		SoundState = 0; // 再生終了
		startPlayback();
		delay(gSound_Length >> 2);		// このdelayの間鳴らしている
		stopPlayback();
	}
//	else if(( gState_F3 > 0) && (gSound_playnext == 2))
	else if((SoundState == 1) && (gSound_playnext == 2))	// DCS50Kは常にファンクッションコマンド送り続けるため、そのガード。
	{
		//Horn (Short)
		gSound_played = 2;
		SoundState = 0; // 再生終了
		startPlayback();
		delay(gSound_Length >> 2);		// このdelayの間鳴らしている
		stopPlayback();
	}
//	else if(( gState_F4 > 0) && (gSound_playnext == 3))
	else if((SoundState == 1) && (gSound_playnext == 3))	// DCS50Kは常にファンクッションコマンド送り続けるため、そのガード。
	{
		//Deactivate break
		gSound_played = 3;
		SoundState = 0; // 再生終了
		
		startPlayback();
		delay(gSound_Length >> 2);		// このdelayの間鳴らしている
		stopPlayback();
	}

}


//DCC速度信号の受信によるイベント
extern void notifyDccSpeed( uint16_t Addr, uint8_t Speed, uint8_t ForwardDir, uint8_t MaxSpeed )
{

	if( gDirection != ForwardDir)
	{
		gDirection = ForwardDir;
	}
	
	gSpeedRef = Speed;
	
// デバッグメッセージ
/*	Serial.print("Speed - ADR: ");
	Serial.print(Addr);
	Serial.print(", SPD: ");
	Serial.print(Speed);
	Serial.print(", DIR: ");
	Serial.print(ForwardDir);
	Serial.print(", MAX: ");
	Serial.println(MaxSpeed);
*/
}

//------------------------------------------------------------------------------
// ファンクション信号受信のイベント
//------------------------------------------------------------------------------
extern void notifyDccFunc( uint16_t Addr, FN_GROUP FuncGrp, uint8_t FuncState)
{
	uint8_t mask;               // 2015/9/12 aya add
	static uint8_t Fa04_f = 0;  // 2015/9/12 aya add

/*
Serial.print("FuncGrp:")   ;Serial.print(FuncGrp);	
Serial.print(" FuncState:");Serial.println(FuncState);
Serial.print("gState_F0:") ;Serial.print(gState_F0);
Serial.print(",gState_F1:");Serial.print(gState_F1);
Serial.print(",gState_F2:");Serial.print(gState_F2);
Serial.print(",gState_F3:");Serial.print(gState_F3);
Serial.print(",gState_F4:");Serial.println(gState_F4);
*/

	if( FuncGrp == FN_0_4)  //Function Group  F0 F4 F3 F2 F1
	{
		if(Fa04_f == 0)		// 初回線路電源ON時に鳴ってしまうガード 2015/9/12 aya add
		{
			Fa04_f = 1;
			pv04_FuncState = FuncState; 
			return;
		}
		mask = pv04_FuncState xor FuncState;      // 前値との変化ビットを算出  2015/9/12 aya add
		pv04_FuncState = FuncState;               // 現在値を保存  2015/9/12 aya add
		
    //Get Function 0 (FL) state
		if( gState_F0 != (mask & FN_BIT_00)){
			gState_F0 = (mask & FN_BIT_00);
		}

    //Get Function 1 state
		if( gState_F1 != (mask & FN_BIT_01)){
      SoundState = 1;   // 再生予約
      gSound_played = 0;
      gSound_playnext = 1;
			gState_F1 = (mask & FN_BIT_01);
		}

    //Get Function 2 state DCS50KではF2ボタンが押されている間、ONになる仕様なので、サウンドが２、３回鳴ってしまいます。
		if( gState_F2 != (mask & FN_BIT_02)){
			SoundState = 1;		// 再生予約
			gSound_played = 0;
			gSound_playnext = 2;
			gState_F2 = (mask & FN_BIT_02);
		}

    //Get Function 3 state
		if( gState_F3 != (mask & FN_BIT_03)){
			SoundState = 1;		// 再生予約
			gSound_played = 0;
			gSound_playnext = 3;
			gState_F3 = (mask & FN_BIT_03);
		}

    //Get Function 4 state
		if( gState_F4 != (mask & FN_BIT_04)) {
			SoundState = 1;		// 再生予約
			gSound_played = 0;
			gSound_playnext = 4;
			gState_F4 = (mask & FN_BIT_04);
		}
	}

}

//------------------------------------------------------------------------------
//  ROMに配置されているサウンドデータから値を取り出す処理
//------------------------------------------------------------------------------
byte readSoundRom(uint8_t inNo, uint16_t inOffset)
{
	switch(inNo)
	{
	case 1:
		return pgm_read_byte(&E233_alert3_data[inOffset]);
		break;
	case 2:
		return pgm_read_byte(&E233_alert1_data[inOffset]);
		break;
	case 3:
		return pgm_read_byte(&E233_break1_data[inOffset]);
		break;
	case 4:
		return 0;
		break;
	default:
		return 0;
		break;
	}
}

void stopPlayback()
{
	// Disable playback per-sample interrupt.
	TIMSK1 &= ~_BV(OCIE1A);

	// Disable the per-sample timer completely.
	TCCR1B &= ~_BV(CS10);

	// Disable the PWM timer.
	TCCR2B &= ~_BV(CS10);

	digitalWrite(speakerPin, LOW);
}


// This is called at 8000 Hz to load the next sample.
ISR(TIMER1_COMPA_vect) 
{
	uint8_t aBufTemp;
	static uint8_t sHiLow = HIGH;						// add aya:最初は上位 IMA ADPCMデータはビッグエンディアンでいいんだよね？

	/* ADPCM先頭ブロック読み出し処理 */
	if( gCntBlockIn == 256 - 4)
	{ 
		// 11/9 1 block 再生終わったら、val,idxを再セットする
		gCntBlockIn = 0;
		gBufIMAdata.val = (int16_t)(readSoundRom(gSound_played,sample + 1) << 8 | readSoundRom(gSound_played,sample)); // Get 1st sample value of the block
		gBufIMAdata.idx = readSoundRom(gSound_played,sample + 2);
		sample += 4;
	} 
    
	if (sample >= gSound_Length) 
	{
		// プチッという音を出さないようにするフェードアウト処理
		if (lastSample == 0) 
		{
			//　音を0まで絞りきったら終了
			stopPlayback();	//　終了処理
		}
		else 
		{
			OCR2B = lastSample;
			lastSample--;				// ボリュームを１つさげる
		}
	}
	else 
	{
    	//ADPCMデコード処理
		aBufTemp = ima_decode(&gBufIMAdata,readSoundRom(gSound_played, sample), sHiLow); // aya add:IMA ADPCM DECODE処理
		OCR2B = aBufTemp;
		
		/* サウンドの最後 */
		if (sample == (gSound_Length - 1))
		{	
			// 最後の音をlastSampleに代入
			lastSample = aBufTemp;
		}
    }

	/* ADPCMの4bit境界切り替え処理 */
	if (sHiLow == HIGH)
	{ 
		// add aya : IMA ADPCM は1byteに4bit*2のデータがあるからね。
		sHiLow = LOW;
	}
	else 
	{
		sHiLow = HIGH;
		sample++;
		gCntBlockIn++;  // 11/9 
	}
}

//------------------------------------------------------------------------------
//  フェードアウト処理
//------------------------------------------------------------------------------
void startPlayback()
{
	pinMode(speakerPin, OUTPUT);

	// Set up Timer 2 to do pulse width modulation on the speaker
	// pin.

	// Use internal clock (datasheet p.160)
	ASSR &= ~(_BV(EXCLK) | _BV(AS2));

	// Set fast PWM mode  (p.157)
	TCCR2A |= _BV(WGM21) | _BV(WGM20);
	TCCR2B &= ~_BV(WGM22);


	// Do non-inverting PWM on pin OC2B (p.155)
	// On the Arduino this is pin 3.
	TCCR2A = (TCCR2A | _BV(COM2B1)) & ~_BV(COM2B0);
	TCCR2A &= ~(_BV(COM2A1) | _BV(COM2A0));
	// No prescaler (p.158)
	TCCR2B = (TCCR2B & ~(_BV(CS12) | _BV(CS11))) | _BV(CS10);

	// Set initial pulse width to the first sample.
	OCR2B = readSoundRom(gSound_played, 0);

	// Set up Timer 1 to send a sample every interrupt.
	cli();

	// Set CTC mode (Clear Timer on Compare Match) (p.133)
	// Have to set OCR1A *after*, otherwise it gets reset to 0!
	TCCR1B = (TCCR1B & ~_BV(WGM13)) | _BV(WGM12);
	TCCR1A = TCCR1A & ~(_BV(WGM11) | _BV(WGM10));

	// No prescaler (p.134)
	TCCR1B = (TCCR1B & ~(_BV(CS12) | _BV(CS11))) | _BV(CS10);

	// Set the compare register (OCR1A).
	// OCR1A is a 16-bit register, so we have to do this with
	// interrupts disabled to be safe.
	OCR1A = F_CPU / SAMPLE_RATE;    // 16e6 / 8000 = 2000

	// Enable interrupt when TCNT1 == OCR1A (p.136)
	TIMSK1 |= _BV(OCIE1A);

	sample = 64;		// dataまでのオフセット
	gCntBlockIn = 0;      // 11/9 
	gSound_Length = (uint16_t)(readSoundRom(gSound_played, 57) <<8 | readSoundRom(gSound_played, 56)); // RIFF No56,No57
	gBufIMAdata.val = (int16_t)(readSoundRom(gSound_played, 61) << 8 | readSoundRom(gSound_played, 60)); // Get 1st sample value of the block
	gBufIMAdata.idx = readSoundRom(gSound_played, 62);
	
	sei();
}
