
#include <stdint.h>
#include <Arduino.h>
#include <avr/interrupt.h>
#include <avr/io.h>
#include "imaadpcm.h"


//------------------------------------------------------------------------------------
// ChaN IMA ADPCM decoder 𗬗p

/*-------------------------------------------------*/
/* IMA ADPCM Decoder                               */

int16_t ima_decode(  /* Returns the sample value (-32768 to 32767) */
  IMADEC* ad,     /* Work area for the ADPCM stream */
  uint8_t dat,    /* ADPCM data, b3:sign (0:+, 1:-), b2-b0:magnify (0..7) */
  uint8_t hl      /* High or low data read */
)
{

#ifdef __PGMSPACE_H_
  static const uint16_t step[89] __attribute__((__progmem__)) = {
#else
  static const uint16_t step[89] = {
#endif
    7, 8, 9, 10, 11, 12, 13, 14, 16, 17, 19, 21, 23, 25, 28, 31, 34, 37, 41, 45, 50, 55, 60, 66,
    73, 80, 88, 97,107, 118, 130, 143, 157, 173, 190, 209, 230, 253, 279, 307, 337, 371, 408, 449,
    494, 544, 598, 658, 724, 796, 876, 963, 1060, 1166, 1282, 1411, 1552, 1707, 1878, 2066, 2272,
    2499, 2749, 3024, 3327, 3660, 4026, 4428, 4871, 5358, 5894, 6484, 7132, 7845, 8630, 9493,
    10442, 11487, 12635, 13899, 15289, 16818, 18500, 20350, 22385, 24623, 27086, 29794, 32767
  };
	int32_t dif;
	int32_t dlevel = 0,adpcmdata;
	int16_t  scale;
	uint16_t stp;
	uint8_t idx;

	// hl ̃tOŁAHIGHA4bitʂɈړ
	if (hl == LOW)
	{
		dat=dat >> 4;
	}
	
	dat = dat & 0x0f;
	
	/*
	sprintf( buffer,"%d,",dat);
	Serial.print(buffer);
	*/
	/* Get current step size */
	idx = ad->idx;            // scale

#ifdef __PGMSPACE_H_
  stp = pgm_read_word(&step[idx]);
#else
  stp = step[idx];
#endif

	/* difference = magnify * (step * 2 + 1) / 8 (calculate without multiplication) */
	dif = stp >> 3;   // dif = stp / 8 Ɠ
	if (dat & 4) dif += stp;
	stp >>= 1;
	if (dat & 2) dif += stp;
	stp >>= 1;
	if (dat & 1) dif += stp;
	if (dat & 8) dif = 1 - dif;
	/* Add the difference to the previous sample */
	dif += ad->val;

	if (dif >  32767) dif =  32767;
	if (dif < -32768) dif = -32768;
	ad->val = (int16_t)dif;

	/* Update step size by magnify */
	dat = (dat & 7) - 4;

	if (dat > 3) {  /* Decrease step size by 1 if magnify is 0..3 */
	if (idx) idx--;
	} else {    /* Increase step size by 2..8 if magnify is 4..7 */
	idx += (dat + 1) << 1;
	if (idx > 88) idx = 88;
	}
	ad->idx = idx;

	return (ad->val >> 8) + 128; /* PCM16bit̃f[^8bitɊԈāA128̃ItZbg𗚂Ă܂*/
}
//------------------------------------------------------------------------------------